import unittest
import array

import numpy as np

import sys
sys.path.append('../../')
import esa_snappy

JAI = esa_snappy.jpy.get_type('javax.media.jai.JAI')
JAI.getDefaultInstance().getTileCache().setMemoryCapacity(128 * 1000 * 1000)

#test_product_file = './MER_RR__1P.N1'
test_product_file = '../testdata/MER_FRS_L1B_SUBSET.dim'

class TestBeamIO(unittest.TestCase):

    def setUp(self):
        self.product = esa_snappy.ProductIO.readProduct(test_product_file)
        self.assertIsNotNone(self.product)


    def tearDown(self):
        self.product.dispose()


    def test_getProductReader(self):
        #print('Band.mro =', snappy.Band.mro())
        reader = self.product.getProductReader()
        self.assertIsNotNone(reader)
        # TODO: fix me: AttributeError: 'org.esa.snap.core.dataio.ProductReader' object has no attribute 'getClass'
        #print('ProductReader.mro =', type(reader).mro())
        #readerClass = reader.getClass()
        #self.assertEqual(readerClass.getName(), '??')


    def test_getBandNames(self):
        names = self.product.getBandNames()
        self.assertEqual(len(names), 17)
        self.assertEqual(names[0], 'radiance_1')
        self.assertEqual(names[14], 'radiance_15')
        self.assertEqual(names[15], 'l1_flags')
        self.assertEqual(names[16], 'detector_index')


    def test_getSceneRasterWidthAndHeight(self):
        w = self.product.getSceneRasterWidth()
        h = self.product.getSceneRasterHeight()
        self.assertEqual(w, 221)
        self.assertTrue(h > 0)


    def test_readPixels_with_java_array(self):
        w = self.product.getSceneRasterWidth()
        h = self.product.getSceneRasterHeight()
        b = self.product.getBand('radiance_13')
        a = esa_snappy.jpy.array('float', w)
        b.readPixels(0, 0, w, 1, a)
        self.assertAlmostEqual(53.841, a[0], 3)
        self.assertTrue(50 < a[100] < 60)
        self.assertAlmostEqual(54.384979, a[100], 6)


    def test_readPixels_with_python_array(self):
        if sys.version_info < (3, 0,):
            # Test only on Python 3.x, as the 2.x array type does not support the new buffer interface
            return
            
        w = self.product.getSceneRasterWidth()
        h = self.product.getSceneRasterHeight()
        b = self.product.getBand('radiance_13')
        a = array.array('f', w * [0])
        b.readPixels(0, 0, w, 1, a)
        self.assertAlmostEqual(53.841, a[0], 3)
        self.assertTrue(50 < a[100] < 60)
        self.assertAlmostEqual(54.384979, a[100], 6)


    def test_readPixels_with_numpy_array(self):
        w = self.product.getSceneRasterWidth()
        h = self.product.getSceneRasterHeight()
        b = self.product.getBand('radiance_13')
        a = np.zeros(w, dtype=np.float32)
        b.readPixels(0, 0, w, 1, a)
        self.assertAlmostEqual(53.841, a[0], 3)
        self.assertTrue(50 < a[100] < 60)
        self.assertAlmostEqual(54.384979, a[100], 6)


    def test_readValidMask_with_numpy_array(self):
        w = self.product.getSceneRasterWidth()
        h = self.product.getSceneRasterHeight()
        b = self.product.getBand('radiance_13')
        a = np.zeros(w, dtype=np.int8)
        #snappy.jpy.diag.flags = snappy.jpy.diag.F_ALL
        b.readValidMask(0, 0, w, 1, a)
        #snappy.jpy.diag.flags = snappy.jpy.diag.F_OFF
        self.assertEqual(a[0], 1)
        self.assertEqual(a[100], 1)


if __name__ == '__main__':
    unittest.main()



import lxml.etree as etree
from .aggregator_avg import AggregatorAvg
from .aggregator_minmax import AggregatorMinMax
from .aggregator_on_max_set import AggregatorOnMaxSet
from .aggregator_percentile import AggregatorPercentile
from .aggregator_avg_outlier import AggregatorAvgOutlier
from .aggregator_sum import AggregatorSum


class Aggregators(object):
    def __init__(self, output_aggregators):

        self.output_aggregators = output_aggregators

    def to_xml(self):

        root = etree.Element("aggregators")

        for aggregator in self.output_aggregators:

            root.append(aggregator.to_xml())

        return root



from esa_snappy import GPF


class OperatorParams:
    def __init__(self, operator):

        self.operator = operator

        self.params = self.get_operator_default_parameters()

    def _get_snap_parameters(self):
        """This function returns the SNAP operator ParameterDescriptors (snappy method op_spi.getOperatorDescriptor().getParameterDescriptors())

        Args:
            operator: SNAP operator

        Returns
            The snappy object returned by op_spi.getOperatorDescriptor().getParameterDescriptors().

        Raises:
            None.
        """
        op_spi = GPF.getDefaultInstance().getOperatorSpiRegistry().getOperatorSpi(self.operator)

        op_params = op_spi.getOperatorDescriptor().getParameterDescriptors()

        return op_params

    def get_operator_default_parameters(self):
        """This function returns a Python dictionary with the SNAP operator parameters and their default values, if available.

        Args:
            operator: SNAP operator

        Returns
            A Python dictionary with the SNAP operator parameters and their default values.

        Raises:
            None.
        """
        parameters = dict()

        for param in self._get_snap_parameters():

            parameters[param.getName()] = param.getDefaultValue()

        return parameters



from attrs import asdict, define, field
import lxml.etree as etree


@define
class AggregatorPercentile(object):

    type = field(init=False, default="PERCENTILE")
    var_name = field()
    target_name = field()
    percentage = field(default=90)

    @percentage.validator
    def _check_type(self, attribute, value):

        if not (isinstance(value, int)):
            raise ValueError("percentage is an int")

        if int(value) < 0 or int(value > 100):
            raise ValueError("percentage value is in range [0.0, 100.0]")

    def __str__(self):

        self.type = "PERCENTILE"

        return self.__repr__()

    def __repr__(self):

        self.type = "PERCENTILE"

        return "AggregatorPercentile({})".format(
            ", ".join(
                ["{}='{}'".format(key, value) for key, value in self.to_dict().items()]
            )
        )

    def to_dict(self):

        self.type = "PERCENTILE"

        return asdict(self)

    def to_xml(self):

        self.type = "PERCENTILE"

        root = etree.Element("aggregator")

        for key, value in self.to_dict().items():

            elem = etree.SubElement(root, key)

            elem.text = str(value)

        return root



#!/usr/bin/env python3

import sys

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')

from esa_snappy import (ProgressMonitor, VectorDataNode,
                        WKTReader, ProductIO, PlainFeatureFactory,
                        SimpleFeatureBuilder, DefaultGeographicCRS,
                        ListFeatureCollection, FeatureUtils)

if len(sys.argv) != 3:
    print("usage: %s <inputProduct> <wkt>" % sys.argv[0])
    sys.exit(1)

input = sys.argv[1]
inputFileName = input[input.rfind('/')+1:]

wellKnownText = sys.argv[2]

try:
    geometry = WKTReader().read(wellKnownText)
except:
    geometry = None
    print('Failed to convert WKT into geometry')
    sys.exit(2)

product = ProductIO.readProduct(input)

wktFeatureType = PlainFeatureFactory.createDefaultFeatureType(DefaultGeographicCRS.WGS84)
featureBuilder = SimpleFeatureBuilder(wktFeatureType)
wktFeature = featureBuilder.buildFeature('shape')
wktFeature.setDefaultGeometry(geometry)

newCollection = ListFeatureCollection(wktFeatureType)
newCollection.add(wktFeature)

productFeatures = FeatureUtils.clipFeatureCollectionToProductBounds(newCollection, product, None, ProgressMonitor.NULL)

node = VectorDataNode('shape', productFeatures)
print ('Num features = ', node.getFeatureCollection().size())

product.getVectorDataGroup().add(node)

vdGroup = product.getVectorDataGroup()
for i in range(vdGroup.getNodeCount()):
    print('Vector data = ', vdGroup.get(i).getName())

maskGroup = product.getMaskGroup()
for i in range(maskGroup.getNodeCount()):
    print('Mask = ', maskGroup.get(i).getName())







import lxml.etree as etree


class TargetBandDescriptors(object):
    
    def __init__(self, target_bands):
        
        self.target_bands = target_bands
        
    def to_xml(self):
        
        root = etree.Element("targetBands")
        
        for target_band in self.target_bands:
            
            root.append(target_band.to_xml())
        
        return root
        


import numpy
import sys

# This example requires a MERIS L2 water product as input

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')

from esa_snappy import Product
from esa_snappy import ProductData
from esa_snappy import ProductIO
from esa_snappy import ProductUtils
from esa_snappy import String


if len(sys.argv) != 2:
    print("usage: %s <file>" % sys.argv[0]);
    sys.exit(1)

print("Reading...")
sourceProduct = ProductIO.readProduct(sys.argv[1])
b1 = sourceProduct.getBand('reflec_5')
b2 = sourceProduct.getBand('reflec_7')
b3 = sourceProduct.getBand('reflec_9')
w1 = b1.getSpectralWavelength()
w2 = b2.getSpectralWavelength()
w3 = b3.getSpectralWavelength()
a = (w2 - w1) / (w3 - w1)
k = 1.03

width = sourceProduct.getSceneRasterWidth()
height = sourceProduct.getSceneRasterHeight()
targetProduct = Product('FLH_Product', 'FLH_Type', width, height)
targetBand = targetProduct.addBand('FLH', ProductData.TYPE_FLOAT32)
ProductUtils.copyGeoCoding(sourceProduct, targetProduct)
targetProduct.setProductWriter(ProductIO.getProductWriter('GeoTIFF'))

targetProduct.writeHeader(String('snappy_flh_output.tif'))

r1 = numpy.zeros(width, dtype=numpy.float32)
r2 = numpy.zeros(width, dtype=numpy.float32)
r3 = numpy.zeros(width, dtype=numpy.float32)

print("Writing...")

for y in range(height):
    b1.readPixels(0, y, width, 1, r1)
    b2.readPixels(0, y, width, 1, r2)
    b2.readPixels(0, y, width, 1, r3)
    print("processing line ", y, " of ", height)
    FLH = r2 - k * (r1 + a * (r3 - r1))
    targetBand.writePixels(0, y, width, 1, FLH)

targetProduct.closeIO()

print("Done.")


from attrs import asdict, define, field
import lxml.etree as etree


@define
class BinningVariable(object):

    name = field()
    expression = field()
    valid_expression = field()

    def __str__(self):

        return self.__repr__()

    def __repr__(self):

        return "BinningVariable({})".format(
            ", ".join(
                ["{}='{}'".format(key, value) for key, value in self.to_dict().items()]
            )
        )

    def to_dict(self):

        return asdict(self)

    def format_key(self, key):

        if key in ["expression"]:

            return "expr"

        elif key in ["valid_expression"]:

            return "validExpr"

        else:

            return key

    def to_xml(self):

        root = etree.Element("variable")

        for key, value in self.to_dict().items():

            elem = etree.SubElement(root, self.format_key(key))

            elem.text = value

        return root



import lxml.etree as etree
import requests 
from urllib.parse import urlparse
from .graph import Graph

def read_file(uri):

    parsed = urlparse(uri)

    if parsed.scheme[:4] == 'http':
    
        r = requests.get(uri)
        root = etree.fromstring(r.content).xpath('/graph')[0]

    else:

        root = etree.parse(uri)

    return Graph(root=root) 


import unittest

import numpy as np

import sys
sys.path.append('../../')
import esa_snappy


#JAI = snappy.jpy.get_type('javax.media.jai.JAI')
#JAI.getDefaultInstance().getTileCache().setMemoryCapacity(256 * 1000 * 1000)

#test_product_file = './MER_RR__1P.N1'
test_product_file = '../testdata/MER_FRS_L1B_SUBSET.dim'

class TestBeamIO(unittest.TestCase):

    def test_readProduct_and_readPixels_mem(self):
        N = 365
        print('test_readProduct_and_readPixels_mem')
        for i in range(N):
            print('i = ', i)
            product = esa_snappy.ProductIO.readProduct(test_product_file)
            band = product.getBand('radiance_9')
            w = band.getRasterWidth()
            h = band.getRasterHeight()
            a = np.zeros(w, dtype=np.float32)
            for y in range(h):
                band.readPixels(0, 0, w, 1, a)
            product.dispose()


if __name__ == '__main__':
    unittest.main()



"""
The esa_snappy module provides access to the Java SE APIs and SNAP Java APIs.

You can configure esa_snappy by using a file named snappy.ini as follows:

    [DEFAULT]
    snap_home: C:/Program Files/snap
    java_class_path: target/classes
    java_library_path: target/lib/amd64
    java_options: -Djava.awt.headless=true
    java_max_mem: 4G
    debug: False

You can place snappy.ini next to <python3>/site-packages/snappy.py or put it in your current working directory.
The 'snap_home' options and all options starting with 'java_' are only used if you use the SNAP API from Python
and a new Java Virtual Machine is created. They are ignored if the SNAP API is called from SNAP itself
(e.g. SNAP command-line or GUI).

'esa_snappy' uses a bundled 'jpy' module, see documentation at http://jpy.readthedocs.org/en/latest
and source code at https://github.com/bcdev/jpy
"""

EXCLUDED_NB_CLUSTERS = {'platform', 'ide', 'bin', 'etc'}

EXCLUDED_DIR_NAMES = {'org.esa.snap.snap-worldwind', 'org.esa.snap.snap-rcp', 'org.esa.snap.snap-product-library',
                      'org.esa.snap.snap-sta-ui'}

EXCLUDED_JAR_NAMES = {'org-esa-snap-netbeans-docwin.jar', 'org-esa-snap-netbeans-tile.jar',
                      'org-esa-snap-snap-worldwind.jar', 'org-esa-snap-snap-tango.jar', 'org-esa-snap-snap-rcp.jar',
                      'org-esa-snap-snap-ui.jar', 'org-esa-snap-snap-graph-builder.jar',
                      'org-esa-snap-snap-branding.jar'}

import glob
import os

import sys

if sys.version_info >= (3, 0, 0,):
    # noinspection PyUnresolvedReferences
    import configparser as cp
else:
    # noinspection PyUnresolvedReferences
    import ConfigParser as cp

module_dir = os.path.dirname(os.path.realpath(__file__))
module_ini = os.path.basename(module_dir) + '.ini'

# Read configuration *.ini file from either '.', '<module_dir>/..', '<module_dir>' in this order
config = cp.ConfigParser()
config.read([os.path.join(module_dir, module_ini),
             os.path.join(os.path.join(module_dir, '..'), module_ini),
             module_ini
             ])

debug = False
if config.has_option('DEFAULT', 'debug'):
    debug = config.getboolean('DEFAULT', 'debug')

# Pre-load Java VM shared library and import 'jpy', the Java-Python bridge
if module_dir not in sys.path:
    sys.path.append(module_dir)
import jpyutil

jpyutil.preload_jvm_dll()
import jpy

if debug:
    # jpy.diag.F_OFF    0x00
    # jpy.diag.F_TYPE   0x01
    # jpy.diag.F_METH   0x02
    # jpy.diag.F_EXEC   0x04
    # jpy.diag.F_MEM    0x08
    # jpy.diag.F_JVM    0x10
    # jpy.diag.F_ERR    0x20
    # jpy.diag.F_ALL    0xff
    jpy.diag.flags = jpy.diag.F_EXEC + jpy.diag.F_ERR


#
# Recursively searches for JAR files in 'dir_path' and appends them to the first member of the
# 'env' tuple (env[0] is the classpath). Also searches for 'lib/LIB_PLATFORM_NAMES[i]' and if one
# exists, appends it to the second member of the 'env' tuple (env[1] is the java.library.path)
# Note: Sub-directories named 'locale' or 'docs' are not searched and
#       JAR files ending with '-ui.jar' or '-examples.jar' are excluded.
#
# Note: This function is called only if the 'esa_snappy' module is not imported from Java.
#
def _collect_snap_jvm_env(dir_path, env):
    for name in os.listdir(dir_path):
        path = os.path.join(dir_path, name)
        if os.path.isfile(path) and name.endswith('.jar'):
            if not (name.endswith('-ui.jar') or name in EXCLUDED_JAR_NAMES):
                env[0][name] = path
        elif os.path.isdir(path) and name not in EXCLUDED_DIR_NAMES:
            if name == 'lib':
                import platform

                os_arch = platform.machine().lower()
                os_name = platform.system().lower()
                lib_os_arch_path = os.path.join(path, os_arch)
                if os.path.exists(lib_os_arch_path):
                    lib_os_name_path = os.path.join(lib_os_arch_path, os_name)
                    if os.path.exists(lib_os_name_path):
                        env[1].append(lib_os_name_path)
                    env[1].append(lib_os_arch_path)
                env[1].append(path)
            if not (name == 'locale' or name == 'docs'):
                _collect_snap_jvm_env(path, env)


#
# Get the NetBeans user directory for installed extra modules
#
def _get_nb_user_modules_dir():
    import platform
    from os.path import expanduser

    home_dir = expanduser('~')
    nb_user_dir = None
    if platform.system() == 'Windows':
        if home_dir:
            nb_user_dir = os.path.join(home_dir, 'AppData\\Roaming\\SNAP')
    else:
        if home_dir:
            nb_user_dir = os.path.join(home_dir, '.snap/system')

    if nb_user_dir:
        return os.path.join(nb_user_dir, 'modules')

    return None


#
# Searches for *.jar files in directory given by global 'snap_home' variable and returns them as a list.
#
# Note: This function is called only if the 'esa_snappy' module is not imported from Java.
#
def _get_snap_jvm_env():
    dir_names = []
    for name in os.listdir(snap_home):
        if os.path.isdir(os.path.join(snap_home, name)):
            dir_names.append(name)

    java_module_dirs = []

    if 'bin' in dir_names and 'etc' in dir_names and 'snap' in dir_names:
        # SNAP Desktop Distribution Directory
        for dir_name in dir_names:
            if dir_name not in EXCLUDED_NB_CLUSTERS:
                dir_path = os.path.join(snap_home, dir_name, 'modules')
                if os.path.isdir(dir_path):
                    java_module_dirs.append(dir_path)
    elif 'lib' in dir_names and 'modules' in dir_names:
        # SNAP Engine Distribution Directory
        java_module_dirs = [os.path.join(snap_home, 'modules'), os.path.join(snap_home, 'lib')]
    elif glob.glob(snap_home + '/*snap-python*.jar'):
        java_module_dirs = [snap_home]
    else:
        raise RuntimeError('does not seem to be a valid SNAP distribution directory: ' + snap_home)

    # NetBeans modules dir will be scaned as last. It contains the latest module updates and they shall replace
    # older modules
    nb_user_modules_dir = _get_nb_user_modules_dir()
    if nb_user_modules_dir and os.path.isdir(nb_user_modules_dir):
        java_module_dirs.append(nb_user_modules_dir)

    if debug:
        import pprint

        print(module_dir + ': java_module_dirs = ')
        pprint.pprint(java_module_dirs)

    env = (dict(), [])
    for path in java_module_dirs:
        _collect_snap_jvm_env(path, env)

    if debug:
        import pprint

        print(module_dir + ': env =')
        pprint.pprint(env)

    return env


#
# Creates a list of Java JVM options, including the Java classpath derived from the global 'snap_home' variable.
#
# Note: This function is called only if the 'esa_snappy' module is not imported from Java.
#
def _get_snap_jvm_options():
    global snap_home

    if config.has_option('DEFAULT', 'snap_home'):
        snap_home = config.get('DEFAULT', 'snap_home')
    else:
        snap_home = os.getenv('SNAP_HOME')

    if snap_home is None or not os.path.isdir(snap_home):
        raise IOError("Can't find SNAP distribution directory. Either configure variable 'snap_home' " +
                      "in file './snappy.ini' or set environment variable 'SNAP_HOME' to an " +
                      "existing SNAP distribution directory.")

    env = _get_snap_jvm_env()
    class_path = list(env[0].values())
    library_path = env[1]

    if config.has_option('DEFAULT', 'java_class_path'):
        extra_class_path = config.get('DEFAULT', 'java_class_path')
        class_path += extra_class_path.split(os.pathsep)

    if config.has_option('DEFAULT', 'java_library_path'):
        extra_library_path = config.get('DEFAULT', 'java_library_path')
        library_path += extra_library_path.split(os.pathsep)

    max_mem = '512M'
    if config.has_option('DEFAULT', 'java_max_mem'):
        max_mem = config.get('DEFAULT', 'java_max_mem')

    options = ['-Djava.awt.headless=true',
               '-Djava.class.path=' + os.pathsep.join(class_path),
               '-Djava.library.path=' + os.pathsep.join(library_path),
               '-Dsnap.home=' + snap_home,
               '-Xmx' + max_mem]

    if config.has_option('DEFAULT', 'java_options'):
        extra_options = config.get('DEFAULT', 'java_options')
        options += extra_options.split('|')

    return options


# Figure out if this module is called from a Java VM. If not, derive a list of Java VM options and create the Java VM.
called_from_java = jpy.has_jvm()
if not called_from_java:
    jpy.create_jvm(options=_get_snap_jvm_options())


# Don't need these functions anymore
del _get_snap_jvm_options
del _get_snap_jvm_env
del _collect_snap_jvm_env


# noinspection PyUnusedLocal
def annotate_RasterDataNode_methods(type_name, method):
    index = -1

    if sys.version_info >= (3, 0, 0,):
        arr_z_type_str = "<class '[Z'>"
        arr_i_type_str = "<class '[I'>"
        arr_f_type_str = "<class '[F'>"
        arr_d_type_str = "<class '[D'>"
    else:
        arr_z_type_str = "<type '[Z'>"
        arr_i_type_str = "<type '[I'>"
        arr_f_type_str = "<type '[F'>"
        arr_d_type_str = "<type '[D'>"

    if method.name == 'readPixels' and method.param_count >= 5:
        index = 4
        param_type_str = str(method.get_param_type(index))
        if param_type_str == arr_i_type_str \
                or param_type_str == arr_f_type_str \
                or param_type_str == arr_d_type_str:
            method.set_param_mutable(index, True)
            method.set_param_output(index, True)
            method.set_param_return(index, True)

    if method.name == 'readValidMask' and method.param_count == 5:
        index = 4
        param_type_str = str(method.get_param_type(index))
        if param_type_str == arr_z_type_str:
            method.set_param_mutable(index, True)
            method.set_param_output(index, True)
            method.set_param_return(index, True)

    if index >= 0 and debug:
        print('annotate_RasterDataNode_methods: Method "{0}": '
              'modified parameter {1:d}: mutable = {2}, return = {3}'
              .format(method.name, index, method.is_param_mutable(index), method.is_param_return(index)))

    return True


jpy.type_callbacks['org.esa.snap.core.datamodel.RasterDataNode'] = annotate_RasterDataNode_methods
jpy.type_callbacks['org.esa.snap.core.datamodel.AbstractBand'] = annotate_RasterDataNode_methods
jpy.type_callbacks['org.esa.snap.core.datamodel.Band'] = annotate_RasterDataNode_methods
jpy.type_callbacks['org.esa.snap.core.datamodel.VirtualBand'] = annotate_RasterDataNode_methods

#
# Preload and assign frequently used Java classes from the Java SE and SNAP Java API.
#

try:
    # Note we may later want to read pre-defined types from a configuration file (snappy.ini)

    # Frequently used classes & interfaces from JRE
    String = jpy.get_type('java.lang.String')
    File = jpy.get_type('java.io.File')
    Point = jpy.get_type('java.awt.Point')
    Rectangle = jpy.get_type('java.awt.Rectangle')
    Arrays = jpy.get_type('java.util.Arrays')
    Collections = jpy.get_type('java.util.Collections')
    List = jpy.get_type('java.util.List')
    Map = jpy.get_type('java.util.Map')
    Set = jpy.get_type('java.util.Set')
    ArrayList = jpy.get_type('java.util.ArrayList')
    HashMap = jpy.get_type('java.util.HashMap')
    HashSet = jpy.get_type('java.util.HashSet')

    #
    # Frequently used classes & interfaces from SNAP Engine
    #

    # Product tree & associates
    Product = jpy.get_type('org.esa.snap.core.datamodel.Product')
    VectorDataNode = jpy.get_type('org.esa.snap.core.datamodel.VectorDataNode')
    RasterDataNode = jpy.get_type('org.esa.snap.core.datamodel.RasterDataNode')
    TiePointGrid = jpy.get_type('org.esa.snap.core.datamodel.TiePointGrid')
    AbstractBand = jpy.get_type('org.esa.snap.core.datamodel.AbstractBand')
    Band = jpy.get_type('org.esa.snap.core.datamodel.Band')
    VirtualBand = jpy.get_type('org.esa.snap.core.datamodel.VirtualBand')
    Mask = jpy.get_type('org.esa.snap.core.datamodel.Mask')
    GeneralFilterBand = jpy.get_type('org.esa.snap.core.datamodel.GeneralFilterBand')
    ConvolutionFilterBand = jpy.get_type('org.esa.snap.core.datamodel.ConvolutionFilterBand')

    # Product tree associates
    ProductData = jpy.get_type('org.esa.snap.core.datamodel.ProductData')
    GeoCoding = jpy.get_type('org.esa.snap.core.datamodel.GeoCoding')
    TiePointGeoCoding = jpy.get_type('org.esa.snap.core.datamodel.TiePointGeoCoding')
    ComponentGeoCoding = jpy.get_type('org.esa.snap.core.dataio.geocoding.ComponentGeoCoding')
    CrsGeoCoding = jpy.get_type('org.esa.snap.core.datamodel.CrsGeoCoding')
    GeoPos = jpy.get_type('org.esa.snap.core.datamodel.GeoPos')
    PixelPos = jpy.get_type('org.esa.snap.core.datamodel.PixelPos')
    FlagCoding = jpy.get_type('org.esa.snap.core.datamodel.FlagCoding')
    ProductNodeGroup = jpy.get_type('org.esa.snap.core.datamodel.ProductNodeGroup')

    # Graph Processing Framework
    GPF = jpy.get_type('org.esa.snap.core.gpf.GPF')
    Operator = jpy.get_type('org.esa.snap.core.gpf.Operator')
    Tile = jpy.get_type('org.esa.snap.core.gpf.Tile')

    # Utilities
    EngineConfig = jpy.get_type('org.esa.snap.runtime.EngineConfig')
    Engine = jpy.get_type('org.esa.snap.runtime.Engine')
    SystemUtils = jpy.get_type('org.esa.snap.core.util.SystemUtils')
    ProductIO = jpy.get_type('org.esa.snap.core.dataio.ProductIO')
    ProductUtils = jpy.get_type('org.esa.snap.core.util.ProductUtils')
    GeoUtils = jpy.get_type('org.esa.snap.core.util.GeoUtils')
    ProgressMonitor = jpy.get_type('com.bc.ceres.core.ProgressMonitor')
    PlainFeatureFactory = jpy.get_type('org.esa.snap.core.datamodel.PlainFeatureFactory')
    FeatureUtils = jpy.get_type('org.esa.snap.core.util.FeatureUtils')

    # GeoTools
    DefaultGeographicCRS = jpy.get_type('org.geotools.referencing.crs.DefaultGeographicCRS')
    ListFeatureCollection = jpy.get_type('org.geotools.data.collection.ListFeatureCollection')
    SimpleFeatureBuilder = jpy.get_type('org.geotools.feature.simple.SimpleFeatureBuilder')

    # JTS
    Geometry = jpy.get_type('org.locationtech.jts.geom.Geometry')
    WKTReader = jpy.get_type('org.locationtech.jts.io.WKTReader')


except Exception:
    jpy.destroy_jvm()
    raise


# Note: use the following code to initialise SNAP's 3rd party libraries, JAI and GeoTools.
# Only needed, if SNAP Python API is not called from Java (e.g. from SNAP gpt or SNAP desktop).
if not called_from_java:
    EngineConfig.instance().load()
    SystemUtils.init3rdPartyLibs(None)

    start_snap_engine = True
    if config.has_option('DEFAULT', 'snap_start_engine'):
        start_snap_engine = config.getboolean('DEFAULT', 'snap_start_engine')

    if start_snap_engine:
        Engine.start()




import argparse
import logging
import os
import os.path
import platform
import sys
import traceback
import zipfile

_ERR_CODE_NO_MATCHING_JPY_WHEEL_FOUND = 10
_ERR_CODE_MISSING_MODULE_JPYUTIL = 20
_ERR_CODE_IMPORTING_SNAPPY_FAILED = 30
_ERR_CODE_IMPORTING_SNAPISTA_FAILED = 40
_ERR_CODE_INTERNAL_ERROR = 50


def _find_file_old(dir_path, regex):
    if os.path.isdir(dir_path):
        for filename in os.listdir(dir_path):
            if regex.match(filename):
                file = os.path.join(dir_path, filename)
                if os.path.isfile(file):
                    return file
    return None


def _find_file(dir_path, jpy_whl_pat):
    if os.path.isdir(dir_path):
        for jpy_whl_filename in os.listdir(dir_path):
            # Windows AMD64
            if "win_amd64" in jpy_whl_pat and "win_amd64" in jpy_whl_filename:
                return _get_file_in_dir(dir_path, jpy_whl_filename)
            # Linux x86_64
            if "linux" in jpy_whl_pat and 'x86_64' in jpy_whl_pat and 'linux' in jpy_whl_filename:
                if 'x86_64' in jpy_whl_filename:
                    return _get_file_in_dir(dir_path, jpy_whl_filename)
            # Linux aarch64
            if 'linux' in jpy_whl_pat and 'aarch64' in jpy_whl_pat and 'linux' in jpy_whl_filename:
                if 'aarch64' in jpy_whl_filename:
                    return _get_file_in_dir(dir_path, jpy_whl_filename)
            # Macos
            if 'macosx' in jpy_whl_pat and 'arm64' in jpy_whl_pat and 'macosx' in jpy_whl_filename:
                if 'x86_64' in jpy_whl_filename or 'universal2' in jpy_whl_filename:
                    return _get_file_in_dir(dir_path, jpy_whl_filename)
    return None


def _get_file_in_dir(dir_path, filename):
    file = os.path.join(dir_path, filename)
    if os.path.isfile(file):
        return file
    else:
        return None


def _configure_snappy(snap_home=None,
                      java_module=None,
                      java_home=None,
                      jvm_max_mem=None,
                      req_arch=None,
                      req_java=False,
                      req_py=False,
                      force=False):
    """
    Unzips matching jpy binary distribution from ../lib/jpy.<platform>-<python-version>.zip,
    imports unpacked 'jpyutil' and configures jpy for SNAP.

    :param snap_home: SNAP distribution directory.
    :param java_home: Java home directory. See also Java system property "java.home".
    :param jvm_max_mem: The heap size of the JVM.
    :param req_arch:  Required JVM architecture (amd64, ia86, x86, etc). See Java system property "os.arch".
    :param req_java:  Fail, if configuration of jpy's Java API fails.
    :param req_py:    Fail, if configuration of jpy's Python API fails.
    :param force:     Force overwriting of existing files.
    :return:
    """

    logging.info("Installing from Java module '" + java_module + "'")

    if req_arch:
        req_arch = req_arch.lower()
        act_arch = platform.machine().lower()

        if req_arch != act_arch:
            logging.warning("Architecture requirement possibly not met: "
                            "Python is " + act_arch + " but JVM requires " + req_arch)
        is64 = sys.maxsize > 2 ** 31 - 1
        if is64 and not req_arch in ('amd64', 'ia64', 'x64', 'x86_64'):
            logging.warning("Architecture requirement possibly not met: "
                            "Python is 64 bit but JVM requires " + req_arch)

    snappy_dir = os.path.dirname(os.path.abspath(__file__))
    snappy_ini_file = os.path.join(snappy_dir, 'esa_snappy.ini')

    jpyutil_file = os.path.join(snappy_dir, 'jpyutil.py')
    jpyconfig_java_file = os.path.join(snappy_dir, 'jpyconfig.properties')
    jpyconfig_py_file = os.path.join(snappy_dir, 'jpyconfig.py')

    must_install_jpy = force \
                       or not os.path.exists(jpyutil_file)
    if not must_install_jpy:
        try:
            import jpyutil
        except ImportError:
            must_install_jpy = True

    must_configure_jpy = force \
                         or must_install_jpy \
                         or not (os.path.exists(jpyconfig_java_file) and os.path.exists(jpyconfig_py_file))
    if not must_configure_jpy:
        try:
            import jpyutil

            jpyutil.preload_jvm_dll()
            import jpy
        except:
            must_configure_jpy = True

    must_configure_snappy = force \
                            or must_configure_jpy \
                            or not os.path.exists(snappy_ini_file)

    #
    # If jpy is not installed yet at all, try to do so by using any compatible binary wheel found
    # in ${snappy_dir} or ${java_module}.
    #
    if must_install_jpy:
        logging.info("Installing jpy...")

        # See "Platform compatibility tags"
        # https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/

        # import distutils.util
        # platform_tag = distutils.util.get_platform().replace('-', '_').replace('.', '_')

        import sysconfig  # switch to sysconfig, as distutils is deprecated in Python 3.10
        platform_tag = sysconfig.get_platform().replace('-', '_').replace('.', '_')
        python_tag = 'cp%d%d' % (sys.version_info.major, sys.version_info.minor,)
        jpy_wheel_file_pat = 'jpy-{version}-%s-{abi_tag}-%s.whl' % (python_tag, platform_tag)

        import re

        jpy_wheel_file_re = jpy_wheel_file_pat.replace('{version}', '[^\-]+').replace('{abi_tag}', '[^\-]+')
        jpy_wheel_file_rec = re.compile(jpy_wheel_file_re)

        #
        # Look for jpy platform wheel in <Python install dir>/Lib/site-packages/esa-snappy/lib
        # For the given Python there should be 4 wheels:
        #     win_amd64, manylinux_x_y_x86_64, manylinux_x_y_aarch64, macos_x_y_universal2
        # in the form jpy-{version}-{python_tag}-{abi_tag}-{platform_tag}.whl
        #
        jpy_wheel_file = _find_file(snappy_dir + os.sep + "lib", jpy_wheel_file_re)

        # No, then search for it in the snap-python module
        # SHOULD NO LONGER BE NEEDED!
        # if not jpy_wheel_file:
        #
        #     # Look for
        #     # ${java_module}/lib/jpy-{version}-{python_tag}-{abi_tag}-{platform_tag}.whl
        #     # depending of whether ${java_module} it is a JAR file or directory
        #
        #     if os.path.isfile(java_module):
        #         with zipfile.ZipFile(java_module) as zf:
        #             lib_prefix = 'lib/'
        #             for name in zf.namelist():
        #                 if name.startswith(lib_prefix):
        #                     basename = name[len(lib_prefix):]
        #                     if jpy_wheel_file_rec.match(basename):
        #                         logging.info("Extracting '" + name + "' from '" + java_module + "'")
        #                         jpy_wheel_file = zf.extract(name, snappy_dir)
        #                         break
        #     else:
        #         lib_dir = os.path.join(java_module, 'lib')
        #         jpy_wheel_file = _find_file(lib_dir, jpy_wheel_file_rec)

        if jpy_wheel_file and os.path.exists(jpy_wheel_file):
            logging.info("Unzipping '" + jpy_wheel_file + "'")
            with zipfile.ZipFile(jpy_wheel_file) as zf:
                zf.extractall(snappy_dir)
        else:
            logging.error("\n".join(
                ["The Java-Python bridge 'jpy' is required to run esa_snappy, but no binary 'jpy' wheel for platform",
                 "'" + sysconfig.get_platform() + "' could be found.",
                 "For further information, please check the esa_snappy log file in: ",
                 "   " + snappy_dir + "/snappyutil.log.",
                 "Please also double-check your installation steps with the documentation in: ",
                 "   https://senbox.atlassian.net/wiki/spaces/SNAP/pages/2499051521/Configure+Python+to+use+the+new+SNAP-Python+esa_snappy+interface+SNAP+version+10",
                 "If this does not help, please feel free to report your problem to the SNAP forum. ",
                 ]))

            return _ERR_CODE_NO_MATCHING_JPY_WHEEL_FOUND
    else:
        logging.info("jpy is already installed")

    #
    # If jpy isn't configured yet, do so by executing "jpyutil.py" which will write the runtime configurations:
    # - "jpyconfig.properties" - Configuration for Java about Python (jpy extension module)
    # - "jpyconfig.py" - Configuration for Python about Java (JVM)
    #
    if must_configure_jpy:
        logging.info("Configuring jpy...")
        if os.path.exists(jpyutil_file):
            # Note 'jpyutil.py' has been unpacked by the previous step, so we can safely import it
            import jpyutil

            if not java_home:
                jre_dir = os.path.join(snap_home, 'jre')
                if not os.path.exists(jre_dir):
                    parent = os.path.dirname(jpyutil_file)
                    while parent:
                        jre_dir = os.path.join(parent, 'jre')
                        if os.path.exists(jre_dir):
                            break
                        parent = os.path.dirname(parent)
                if os.path.exists(jre_dir):
                    java_home = os.path.normpath(jre_dir)

            ret_code = jpyutil.write_config_files(out_dir=snappy_dir,
                                                  java_home_dir=java_home,
                                                  req_java_api_conf=req_java,
                                                  req_py_api_conf=req_py)
            if ret_code:
                return ret_code
        else:
            logging.error("Missing Python module '{}'\nwhich is required to complete the configuration."
                          .format(jpyutil_file))
            return _ERR_CODE_MISSING_MODULE_JPYUTIL
    else:
        logging.info("jpy is already configured")

    #
    # If snappy isn't configured yet, do so by writing a default "snappy.ini".
    # Note, this file is only used if you use SNAP from Python, i.e. importing
    # the snappy module in your Python programs.
    #
    if must_configure_snappy:
        logging.info("Configuring snappy...")
        with open(snappy_ini_file, 'w') as file:
            file.writelines(['[DEFAULT]\n',
                             'snap_home = %s\n' % snap_home,
                             'java_max_mem: %s\n' % jvm_max_mem,
                             '# snap_start_engine: False\n',
                             '# java_class_path: ./target/classes\n',
                             '# java_library_path: ./lib\n',
                             '# java_options: -Djava.awt.headless=false\n',
                             '# debug: False\n'])
            logging.info("snappy configuration written to '{}'".format(snappy_ini_file))
    else:
        logging.info("snappy is already configured")

    #
    # Finally, we test the snappy installation/configuration by importing it.
    # If this won't succeed, _main() will catch the error and report it.
    #
    logging.info("Importing esa_snappy.snapista for final test...")
    sys.path = [os.path.join(snappy_dir, '..')] + sys.path

    try:
        __import__('esa_snappy.snapista')
    except:
        return _ERR_CODE_IMPORTING_SNAPISTA_FAILED

    logging.info("Done. The SNAP-Python interface is located as package 'esa_snappy' in "
                 "your Python's 'site-packages' directory..\n")

    return 0


def _main():
    parser = argparse.ArgumentParser(description='Configures snappy, the SNAP-Python interface.')
    parser.add_argument('--snap_home', default=None,
                        help='SNAP distribution directory')
    parser.add_argument('--req_arch', default=None,
                        help='required JVM architecture, e.g. "amd64", '
                             'may be taken from Java system property "os.arch"')
    parser.add_argument('--java_module', default=None,
                        help='directory or JAR file containing the "snap-python" Java module')
    parser.add_argument('--java_home', default=None,
                        help='Java JDK or JRE installation directory, '
                             'may be taken from Java system property "java.home"')
    parser.add_argument('--jvm_max_mem', default='3G', help='size of the Java VM heap space')
    parser.add_argument("--log_file", action='store', default=None, help="file into which to write logging output")
    parser.add_argument("--log_level", action='store', default='INFO',
                        help="log level, possible values are: DEBUG, INFO, WARNING, ERROR")
    parser.add_argument("-j", "--req_java", action='store_true', default=False,
                        help="require that Java API configuration succeeds")
    parser.add_argument("-p", "--req_py", action='store_true', default=False,
                        help="require that Python API configuration succeeds")
    parser.add_argument('-f', '--force', action='store_true', default=False,
                        help='force overwriting of existing files')
    args = parser.parse_args()

    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        raise ValueError('Invalid log level: %s' % log_level)

    log_format = '%(levelname)s: %(message)s'
    log_file = args.log_file
    if log_file:
        logging.basicConfig(format=log_format, level=log_level, filename=log_file, filemode='w')
    else:
        logging.basicConfig(format=log_format, level=log_level)

    # noinspection PyBroadException
    try:
        ret_code = _configure_snappy(snap_home=args.snap_home,
                                     java_module=args.java_module,
                                     java_home=args.java_home,
                                     jvm_max_mem=args.jvm_max_mem,
                                     req_arch=args.req_arch,
                                     req_java=args.req_java,
                                     req_py=args.req_py,
                                     force=args.force)
    except:
        ret_code = _ERR_CODE_INTERNAL_ERROR
        logging.error(traceback.format_exc())

    if ret_code != 0:
        logging.error("Configuration failed with exit code {}".format(ret_code))

    exit(ret_code)


if __name__ == '__main__':
    _main()



import sys

# This example requires a MERIS L1b product as input

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')

from esa_snappy import GPF
from esa_snappy import HashMap
from esa_snappy import ProductIO
from esa_snappy import jpy

if len(sys.argv) != 2:
    print("usage: %s <file>" % sys.argv[0])
    sys.exit(1)

file = sys.argv[1]

print("Reading...")
product = ProductIO.readProduct(file)
width = product.getSceneRasterWidth()
height = product.getSceneRasterHeight()
name = product.getName()
description = product.getDescription()
band_names = product.getBandNames()

print("Product: %s, %d x %d pixels, %s" % (name, width, height, description))
print("Bands:   %s" % (list(band_names)))

GPF.getDefaultInstance().getOperatorSpiRegistry().loadOperatorSpis()

BandDescriptor = jpy.get_type('org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor')

targetBand1 = BandDescriptor()
targetBand1.name = 'band_1'
targetBand1.type = 'float32'
targetBand1.expression = '(radiance_10 - radiance_7) / (radiance_10 + radiance_7)'

targetBand2 = BandDescriptor()
targetBand2.name = 'band_2'
targetBand2.type = 'float32'
targetBand2.expression = '(radiance_9 - radiance_6) / (radiance_9 + radiance_6)'

if product.getBand('radiance_10') is None:
    # probably 4RP
    targetBand1.expression = '(M10_radiance - M07_radiance) / (M10_radiance + M07_radiance)'
    targetBand2.expression = '(M09_radiance - M06_radiance) / (M09_radiance + M06_radiance)'

targetBands = jpy.array('org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor', 2)
targetBands[0] = targetBand1
targetBands[1] = targetBand2

parameters = HashMap()
parameters.put('targetBands', targetBands)

result = GPF.createProduct('BandMaths', parameters, product)

print("Writing...")

ProductIO.writeProduct(result, 'snappy_bmaths_output.dim', 'BEAM-DIMAP')

print("Done.")



from types import SimpleNamespace
from esa_snappy import GPF, jpy
from .operatorparams import OperatorParams


class Operator(SimpleNamespace):
    def __init__(self, operator, **kwargs):

        self.operator = operator
        self._params = {**OperatorParams(self.operator).params, **kwargs}

        return super().__init__(**self._params)

    def __str__(self):

        return "{}:\n\t{}".format(
            self.operator, "\n\t".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def __repr__(self):

        return "Operator('{}', {})".format(
            self.operator, ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def to_dict(self):

        return dict([(name, getattr(self, name)) for name in list(self._params.keys())])

    def describe(self):
        """This function prints the human readable information about a SNAP operator

        Args:
            operator: SNAP operator

        Returns
            The human readable information about the provided SNAP operator.

        Raises:
            None.
        """
        op_spi = GPF.getDefaultInstance().getOperatorSpiRegistry().getOperatorSpi(self.operator)

        print("Operator name: {}\n".format(op_spi.getOperatorDescriptor().getAlias()))
        print("Description: {}".format(op_spi.getOperatorDescriptor().getDescription()))
        print("Authors: {}\n".format(op_spi.getOperatorDescriptor().getAuthors()))
        print("{}".format(op_spi.getOperatorDescriptor().getName()))
        print("Version: {}\n".format(op_spi.getOperatorDescriptor().getVersion()))
        print("Parameters:\n")
        param_Desc = op_spi.getOperatorDescriptor().getParameterDescriptors()

        for param in param_Desc:
            print(
                "\t{}: {}\n\t\tDefault Value: {}\n".format(
                    param.getName(), param.getDescription(), param.getDefaultValue()
                )
            )

            if self.operator == "Write" and param.getName() == "formatName":

                print("\t\tPossible values: {}\n".format(self._get_formats("Write")))

            elif self.operator == "Read" and param.getName() == "formatName":

                print("\t\tPossible values: {}\n".format(self._get_formats("Read")))

            else:

                print("\t\tPossible values: {}\n".format(list(param.getValueSet())))

    @staticmethod
    def _get_formats(method):
        """This function provides a human readable list of SNAP Read or Write operator formats.

        Args:
            None.

        Returns
            Human readable list of SNAP Write operator formats.

        Raises:
            None.
        """
        ProductIOPlugInManager = jpy.get_type("org.esa.snap.core.dataio.ProductIOPlugInManager")

        if method == "Read":
            plugins = ProductIOPlugInManager.getInstance().getAllReaderPlugIns()
        elif method == "Write":
            plugins = ProductIOPlugInManager.getInstance().getAllWriterPlugIns()
        else:
            raise ValueError

        formats = []

        while plugins.hasNext():
            plugin = plugins.next()
            formats.append(plugin.getFormatNames()[0])

        return formats



import unittest

import numpy as np

import sys
sys.path.append('../../')
import esa_snappy


#JAI = snappy.jpy.get_type('javax.media.jai.JAI')
#JAI.getDefaultInstance().getTileCache().setMemoryCapacity(128 * 1000 * 1000)

#test_product_file = './MER_RR__1P.N1'
test_product_file = '../testdata/MER_FRS_L1B_SUBSET.dim'


class TestBeamIO(unittest.TestCase):

    def setUp(self):
        self.product = esa_snappy.ProductIO.readProduct(test_product_file)
        self.assertIsNotNone(self.product)


    def tearDown(self):
        self.product.dispose()


    def test_readPixels_performance(self):
        w = self.product.getSceneRasterWidth()
        h = self.product.getSceneRasterHeight()
        b = self.product.getBand('radiance_13')
        a = np.zeros(w, dtype=np.float32)

        import time

        t0 = time.time()
        for y in range(h):
            b.readPixels(0, 0, w, 1, a)
        t1 = time.time()

        dt = t1 - t0
        print('Band.readPixels(): w =', w, ', dtype=np.float32:', h, 'calls in', dt*1000, 'ms, that is ', dt*1000/y, 'ms per call')


    def test_readValidMask_performance(self):
        w = self.product.getSceneRasterWidth()
        h = self.product.getSceneRasterHeight()
        b = self.product.getBand('radiance_13')
        a = np.zeros(w, dtype=bool)

        import time

        t0 = time.time()
        for y in range(h):
            b.readValidMask(0, 0, w, 1, a)
        t1 = time.time()

        dt = t1 - t0
        print('Band.readValidMask(): w =', w, ', dtype=bool:', h, 'calls in', dt*1000, 'ms, that is ', dt*1000/y, 'ms per call')


if __name__ == '__main__':
    unittest.main()



from .graph import Graph
from .graph_io import read_file
from .operator import Operator
from .operatorparams import OperatorParams
from .target_band import TargetBand
from .target_band_descriptors import TargetBandDescriptors


import os
import subprocess
import tempfile
import platform
import lxml.etree as etree
from esa_snappy import GPF
from xml.sax.saxutils import unescape
from esa_snappy.snapista.binning.output_bands import BinningOutputBands
from .target_band_descriptors import TargetBandDescriptors
from .binning import Aggregators
from .binning import BinningVariables


class Graph:
    """SNAP Graph class

    This class provides the methods to create, view and run a SNAP Graph

    Attributes:
        None.
    """

    def __init__(self, wdir=".", root=None):

        if root is None:
            self.root = etree.Element("graph")

            version = etree.SubElement(self.root, "version")
            version.text = "1.0"

        else:
            self.root = root

        self.pid = None
        self.p = None
        self.wdir = wdir
        self.gpt_path = self.get_gpt_cmd()

        if not self.gpt_path:

            raise Exception("gpt not found")

    def __str__(self):

        return "gpt binary: {}\nworking dir: {}\n\n{}".format(
            self.gpt_path,
            self.wdir,
            etree.tostring(self.root, pretty_print=True).decode("utf-8"),
        ).replace("\\n", "\n")

    def __repr__(self):

        return "Graph(wdir='{}')".format(self.wdir)

    @staticmethod
    def get_gpt_cmd():

        gpt_cmd = None

        if platform.system() == 'Windows':
            path_split_char = ";"
            gpt_executable = "gpt.exe"
        else:
            path_split_char = ":"
            gpt_executable = "gpt"

        for p in os.getenv("PATH").split(path_split_char):

            if os.path.exists(os.path.join(p, gpt_executable)):

                gpt_cmd = os.path.join(p, gpt_executable)

                break

        return gpt_cmd

    @staticmethod
    def list_operators():
        """This function provides a Python dictionary with all SNAP operators.

        Args:
            None.

        Returns
            Python dictionary with all SNAP operators.

        Raises:
            None.
        """
        GPF.getDefaultInstance().getOperatorSpiRegistry().loadOperatorSpis()

        op_spi_it = (
            GPF.getDefaultInstance()
            .getOperatorSpiRegistry()
            .getOperatorSpis()
            .iterator()
        )

        snap_operators = []

        while op_spi_it.hasNext():

            op_spi = op_spi_it.next()

            snap_operators.append(op_spi.getOperatorAlias())

        return snap_operators

    @staticmethod
    def describe_operators():
        """This function provides a Python dictionary with all SNAP operators.

        Args:
            None.

        Returns
            Python dictionary with all SNAP operators.

        Raises:
            None.
        """

        desc_dict = {}

        GPF.getDefaultInstance().getOperatorSpiRegistry().loadOperatorSpis()

        op_spi_it = (
            GPF.getDefaultInstance()
            .getOperatorSpiRegistry()
            .getOperatorSpis()
            .iterator()
        )

        while op_spi_it.hasNext():

            op_spi = op_spi_it.next()

            alias = op_spi.getOperatorDescriptor().getAlias()
            description = op_spi.getOperatorDescriptor().getDescription()

            desc_dict[alias] = description
            print("{} - {}".format(alias, description))

        return desc_dict

    # todo: seems that this is not needed. Check.
    def nice_view(self):

        try:
            from pygments import highlight
            from pygments.lexers import XmlLexer
            from pygments.formatters import HtmlFormatter
            import IPython
            from IPython.display import HTML

            def display_xml_nice(xml):
                formatter = HtmlFormatter()
                IPython.display.display(
                    HTML(
                        '<style type="text/css">{}</style>    {}'.format(
                            formatter.get_style_defs(".highlight"),
                            highlight(xml, XmlLexer(), formatter),
                        )
                    )
                )

            display_xml_nice(etree.tostring(self.root, pretty_print=True))

        except ModuleNotFoundError:

            print(etree.tostring(self.root, pretty_print=True).decode("utf-8")).replace(
                "\\n", "\n"
            )

    def view(self):
        """This method prints SNAP Graph

        Args:
            None.

        Returns
            None.

        Raises:
            None.
        """
        print(unescape(etree.tostring(self.root, pretty_print=True).decode("utf-8")))

    def add_node(self, operator, node_id, source=None):
        """This method adds or overwrites a node to the SNAP Graph

        Args:
            operator: SNAP operator
            node_id: node identifier
            source: string or list of sources (previous node identifiers in the SNAP Graph)

        Returns
            None.

        Raises:
            None.
        """
        xpath_expr = '/graph/node[@id="%s"]' % node_id

        if len(self.root.xpath(xpath_expr)) != 0:

            node_elem = self.root.xpath(xpath_expr)[0]
            operator_elem = self.root.xpath(xpath_expr + "/operator")[0]
            sources_elem = self.root.xpath(xpath_expr + "/sources")[0]
            parameters_elem = self.root.xpath(xpath_expr + "/parameters")

            for param in [
                name
                for name in dir(operator)
                if name[:2] != "__"
                and name[-2:] != "__"
                and name != "_params"
                and name != "operator"
                and type(getattr(operator, name)).__name__
                in [
                    "str",
                    "NoneType",
                    "TargetBandDescriptors",
                    "Aggregators",
                    "BinningOutputBands",
                    "BinningVariables",
                ]
            ]:
              
                if param in [
                    "targetBandDescriptors",
                    "aggregatorConfigs",
                    "variableConfigs",
                    "bandConfigurations",
                    "postProcessorConfig",
                    "productCustomizerConfig",
                ]:
                  
                    if (param in ["bandConfigurations", "variableConfigs", "postProcessorConfig", "productCustomizerConfig" ] and not getattr(operator, param)): continue

                    if (
                        isinstance(getattr(operator, param), TargetBandDescriptors)
                        or isinstance(getattr(operator, param), Aggregators)
                        or isinstance(getattr(operator, param), BinningOutputBands)
                        or isinstance(getattr(operator, param), BinningVariables)
                    ):
                        parameters_elem.append(getattr(operator, param).to_xml())

                    elif isinstance(getattr(operator, param), str):

                        parameters_elem.append(
                            etree.fromstring(getattr(operator, param))
                        )
                    else:

                        raise ValueError()

                else:
                    try: 
                        p_elem = self.root.xpath(xpath_expr + "/parameters/%s" % param)[0]

                        if getattr(operator, param) is not None:
                            if getattr(operator, param)[0] != "<":
                                p_elem.text = getattr(operator, param)
                            else:
                                p_elem.text.append(
                                    etree.fromstring(getattr(operator, param))
                                )
                    except IndexError:
                        pass

        else:

            node_elem = etree.SubElement(self.root, "node")
            operator_elem = etree.SubElement(node_elem, "operator")
            sources_elem = etree.SubElement(node_elem, "sources")

            if isinstance(source, list):

                for index, s in enumerate(source):
                    if index == 0:
                        source_product_elem = etree.SubElement(
                            sources_elem, "sourceProduct"
                        )

                    else:
                        source_product_elem = etree.SubElement(
                            sources_elem, "sourceProduct.%s" % str(index)
                        )

                    source_product_elem.attrib["refid"] = s

            elif isinstance(source, dict):

                for key, value in source.iteritems():

                    source_product_elem = etree.SubElement(sources_elem, key)
                    source_product_elem.text = value

            elif source is not None:
                source_product_elem = etree.SubElement(sources_elem, "sourceProduct")
                source_product_elem.attrib["refid"] = source

            parameters_elem = etree.SubElement(node_elem, "parameters")
            parameters_elem.attrib["class"] = "com.bc.ceres.binding.dom.XppDomElement"

            for param in [
                name
                for name in dir(operator)
                if name[:2] != "__"
                and name[-2:] != "__"
                and name != "_params"
                and name != "operator"
                and type(getattr(operator, name)).__name__
                in [
                    "str",
                    "NoneType",
                    "TargetBandDescriptors",
                    "Aggregators",
                    "BinningOutputBands",
                    "BinningVariables",
                ]
            ]:

                if param in [
                    "targetBandDescriptors",
                    "aggregatorConfigs",
                    "variableConfigs",
                    "bandConfigurations",
                    "postProcessorConfig",
                    "productCustomizerConfig",
                ]:
                    print(param, getattr(operator, param))
                    if (param in ["bandConfigurations", "variableConfigs", "postProcessorConfig", "productCustomizerConfig" ] and not getattr(operator, param)): continue

                    print('Instance TargetBandDescriptors: ' + str(isinstance(getattr(operator, param), TargetBandDescriptors)))
                    print('Instance Aggregators: ' + str(isinstance(getattr(operator, param), Aggregators)))
                    print('Instance BinningOutputBands: ' + str(isinstance(getattr(operator, param), BinningOutputBands)))
                    print('Instance BinningVariables: ' + str(isinstance(getattr(operator, param), BinningVariables)))
                    print('Instance str: ' + str(isinstance(getattr(operator, param), str)))
                    if (
                        isinstance(getattr(operator, param), TargetBandDescriptors)
                        or isinstance(getattr(operator, param), Aggregators)
                        or isinstance(getattr(operator, param), BinningOutputBands)
                        or isinstance(getattr(operator, param), BinningVariables)
                    ):

                        parameters_elem.append(getattr(operator, param).to_xml())

                    elif isinstance(getattr(operator, param), str):

                        parameters_elem.append(
                            etree.fromstring(getattr(operator, param))
                        )
                    else:

                        raise ValueError()

                else:

                    parameter_elem = etree.SubElement(parameters_elem, param)

                    if getattr(operator, param) is not None:
                        if getattr(operator, param)[0] != "<":
                            parameter_elem.text = getattr(operator, param)
                        else:
                            parameter_elem.append(
                                etree.fromstring(getattr(operator, param))
                            )

        node_elem.attrib["id"] = node_id

        operator_elem.text = operator.operator

    def save_graph(self, filename):
        """This method saves the SNAP Graph

        Args:
            filename: XML filename with '.xml' extension

        Returns
            None.

        Raises:
            None.
        """
        
        with open(filename, "w") as file:
            file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            file.write(unescape(etree.tostring(self.root, pretty_print=True).decode()))

    def run(self, gpt_options=["-x", "-c", "1024M"]):
        """This method runs the SNAP Graph using gpt

        Args:
            gpt_options: list of options to pass to gpt. Defaults to ['-x', '-c', '1024M']

        Returns
            res: gpt exit code
            err: gpt stderr

        Raises:
            None.
        """

        def _run_command(command, **kwargs):

            process = subprocess.Popen(args=command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,  **kwargs)
            while True:
                output = process.stdout.readline()
                err = process.stderr.readline()
                if output.decode() == "" and process.poll() is not None:
                    break
                if output:
                    print(output.strip().decode())
                if err:
                    print(err.strip().decode())
            rc = process.poll()

            return rc

        os.environ["LD_LIBRARY_PATH"] = "."

        print("Processing the graph")

        fd, path = tempfile.mkstemp()

        rc = None

        try:

            self.save_graph(filename=path)

            options = [self.gpt_path, *gpt_options, path]

            rc = _run_command(options)

        finally:

            os.remove(path)

        if rc != 0:

            raise Exception("Graph execution failed (exit code {})".format(rc))

        return rc



import attr
import lxml.etree as etree
from xml.sax.saxutils import escape


@attr.s
class TargetBand(object):

    name = attr.ib()
    expression = attr.ib()
    type = attr.ib('float32')
    description = attr.ib(default=None)
    unit = attr.ib(default=None)
    no_data_value = attr.ib(default="NaN")


    def __str__(self):

        return self.__repr__()

    def __repr__(self):

        return "TargetBand({})".format(
            ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def to_dict(self):
       
        return attr.asdict(self)
    
    def to_xml(self):
        
        root = etree.Element("targetBand")
        
        for key, value in self.to_dict().items():

            elem = etree.SubElement(root, key)

            if key == 'expression':
                elem.text = escape(value)
            else:
                elem.text = value
            
        return root


import sys

import numpy

# This example requires a MERIS L1b product as input

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')

from esa_snappy import Product
from esa_snappy import ProductData
from esa_snappy import ProductIO
from esa_snappy import ProductUtils
from esa_snappy import FlagCoding

if len(sys.argv) != 2:
    print("usage: %s <file>" % sys.argv[0])
    sys.exit(1)

file = sys.argv[1]

print("Reading...")
product = ProductIO.readProduct(file)
width = product.getSceneRasterWidth()
height = product.getSceneRasterHeight()
name = product.getName()
description = product.getDescription()
band_names = product.getBandNames()

print("Product:     %s, %s" % (name, description))
print("Raster size: %d x %d pixels" % (width, height))
print("Start time:  " + str(product.getStartTime()))
print("End time:    " + str(product.getEndTime()))
print("Bands:       %s" % (list(band_names)))


b7 = product.getBand('radiance_7')
b10 = product.getBand('radiance_10')
if b10 is None:
    b7 = product.getBand('M07_radiance')
    b10 = product.getBand('M10_radiance')
ndviProduct = Product('NDVI', 'NDVI', width, height)
ndviBand = ndviProduct.addBand('ndvi', ProductData.TYPE_FLOAT32)
ndviFlagsBand = ndviProduct.addBand('ndvi_flags', ProductData.TYPE_UINT8)
writer = ProductIO.getProductWriter('BEAM-DIMAP')

ProductUtils.copyGeoCoding(product, ndviProduct)

ndviFlagCoding = FlagCoding('ndvi_flags')
ndviFlagCoding.addFlag("NDVI_LOW", 1, "NDVI below 0")
ndviFlagCoding.addFlag("NDVI_HIGH", 2, "NDVI above 1")
group = ndviProduct.getFlagCodingGroup()
#print(dir(group))
group.add(ndviFlagCoding)

ndviFlagsBand.setSampleCoding(ndviFlagCoding)

ndviProduct.setProductWriter(writer)
ndviProduct.writeHeader('snappy_ndvi_output.dim')

r7 = numpy.zeros(width, dtype=numpy.float32)
r10 = numpy.zeros(width, dtype=numpy.float32)

print("Writing...")

for y in range(height):
    print("processing line ", y, " of ", height)
    r7 = b7.readPixels(0, y, width, 1, r7)
    r10 = b10.readPixels(0, y, width, 1, r10)

    ndvi = (r10 - r7) / (r10 + r7)
    ndviBand.writePixels(0, y, width, 1, ndvi)
    #ndviLow = ndvi < 0.0
    #ndviHigh = ndvi > 1.0
    #ndviFlags = numpy.array(ndviLow + 2 * ndviHigh, dtype=numpy.int32)
    #ndviFlagsBand.writePixels(0, y, width, 1, ndviFlags)

ndviProduct.closeIO()

print("Done.")


from attrs import asdict, define, field
import lxml.etree as etree


@define
class BinningBand(object):

    index = field()
    name = field()
    min_value = field()
    max_value = field()

    def __str__(self):

        return self.__repr__()

    def __repr__(self):

        return "BinningBand({})".format(
            ", ".join(
                ["{}='{}'".format(key, value) for key, value in self.to_dict().items()]
            )
        )

    def to_dict(self):

        return asdict(self)

    def to_xml(self):

        root = etree.Element("band")

        for key, value in self.to_dict().items():

            elem = etree.SubElement(root, key)

            elem.text = str(value)

        return root



#import attr
import lxml.etree as etree
from attrs import asdict, define, field


# @attr.s
@define
class AggregatorAvg(object):

    # type = attr.ib(init=False, default="AVG")
    # varName = attr.ib()
    # targetName = attr.ib()
    # weightCoeff = attr.ib(default=0.0)
    # outputCounts = attr.ib(default="true")
    # outputSums = attr.ib(default="true")

    type = field(init=False, default="AVG")
    var_name = field()
    target_name = field()
    weight_coeff = field(default=0.0)
    output_counts = field(default="true")
    output_sums = field(default="true")
    @output_counts.validator
    def _check_type(self, attribute, value):
        if value not in ["true", "false"]:
            raise ValueError("output_counts value is either 'true' or 'false'")

    @output_sums.validator
    def _check_type(self, attribute, value):
        if value not in ["true", "false"]:
            raise ValueError("output_sums value is either 'true' or 'false'")

    def __str__(self):

        self.type = "AVG"

        return self.__repr__()

    def __repr__(self):

        self.type = "AVG"

        return "AggregatorAvg({})".format(
            ", ".join(
                ["{}='{}'".format(key, value) for key, value in self.to_dict().items()]
            )
        )

    def to_dict(self):

        self.type = "AVG"

        # return attr.asdict(self)
        return asdict(self)

    def to_xml(self):

        self.type = "AVG"

        root = etree.Element("aggregator")

        for key, value in self.to_dict().items():

            elem = etree.SubElement(root, key)

            elem.text = str(value)

        return root



import sys

# This example requires a MERIS L1b product as input

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')

import esa_snappy
from esa_snappy import (ProductIO, ProductUtils, ProgressMonitor)

if len(sys.argv) != 2:
    print("usage: %s <file>" % sys.argv[0])
    sys.exit(1)

file = sys.argv[1]

jpy = esa_snappy.jpy

# More Java type definitions required for image generation
Color = jpy.get_type('java.awt.Color')
ColorPoint = jpy.get_type('org.esa.snap.core.datamodel.ColorPaletteDef$Point')
ColorPaletteDef = jpy.get_type('org.esa.snap.core.datamodel.ColorPaletteDef')
ImageInfo = jpy.get_type('org.esa.snap.core.datamodel.ImageInfo')
ImageLegend = jpy.get_type('org.esa.snap.core.datamodel.ImageLegend')
ImageManager = jpy.get_type('org.esa.snap.core.image.ImageManager')
JAI = jpy.get_type('javax.media.jai.JAI')
RenderedImage = jpy.get_type('java.awt.image.RenderedImage')
Integer = jpy.get_type('java.lang.Integer')


# Disable JAI native MediaLib extensions 
System = jpy.get_type('java.lang.System')
System.setProperty('com.sun.media.jai.disableMediaLib', 'true')


def write_image(band, filename, format):
    im = ImageManager.getInstance().createColoredBandImage([band], band.getImageInfo(), 0)
    JAI.create("filestore", im, filename, format)


def write_rgb_image(bands, filename, format):
    image_info = ProductUtils.createImageInfo(bands, True, ProgressMonitor.NULL)
    im = ImageManager.getInstance().createColoredBandImage(bands, image_info, 0)
    JAI.create("filestore", im, filename, format)


def resize(product, targetWidth, targetHeight):
    from esa_snappy import GPF
    from esa_snappy import HashMap

    parameters = HashMap()
    parameters.put('targetWidth', Integer(targetWidth))
    parameters.put('targetHeight', Integer(targetHeight))
    return GPF.createProduct('Resample', parameters, product)

product = ProductIO.readProduct(file)

# This scales the product to the specified size
# remove if you don't want scaling.
product = resize(product, 1000, 1000)

band = product.getBand('radiance_13')
if band is None:
    band = product.getBand('M13_radiance')

# The colour palette assigned to pixel values 0, 50, 100 in the band's geophysical units
points = [ColorPoint(0.0, Color.YELLOW), 
          ColorPoint(50.0, Color.RED), 
          ColorPoint(100.0, Color.BLUE)]
cpd = ColorPaletteDef(points)
ii = ImageInfo(cpd)
band.setImageInfo(ii)

image_format = 'PNG'
write_image(band, 'snappy_write_image.png', image_format)

legend = ImageLegend(band.getImageInfo(), band)
legend.setHeaderText(band.getName())

#legend.setOrientation(ImageLegend.HORIZONTAL) # or ImageLegend.VERTICAL
#legend.setFont(legend.getFont().deriveFont(14))
#legend.setBackgroundColor(Color.CYAN)
#legend.setForegroundColor(Color.ORANGE);
#legend.setBackgroundTransparency(0.7);
#legend.setBackgroundTransparencyEnabled(True);
#legend.setAntialiasing(True);

legend_image = legend.createImage()

# This cast is need because otherwise jpy can't evaluate which method to call
# This is considered as an issue of jpy (https://github.com/bcdev/jpy/issues/89)
rendered_legend_image = jpy.cast(legend_image, RenderedImage)
JAI.create("filestore", rendered_legend_image, 'snappy_write_image_legend.png', image_format)

red = product.getBand('radiance_13')
green = product.getBand('radiance_5')
blue = product.getBand('radiance_1')
if red is None:
    red = product.getBand('M13_radiance')
    green = product.getBand('M05_radiance')
    blue = product.getBand('M01_radiance')
write_rgb_image([red, green, blue], 'snappy_write_image_rgb.png', image_format)







import sys

import numpy

# This example requires a MERIS L1b product as input

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')

from esa_snappy import String
from esa_snappy import Product
from esa_snappy import ProductData
from esa_snappy import ProductIO
from esa_snappy import ProductUtils


if len(sys.argv) != 2:
    print("usage: %s <file>" % sys.argv[0]);
    sys.exit(1)

print("Reading...")
product = ProductIO.readProduct(sys.argv[1])
width = product.getSceneRasterWidth()
height = product.getSceneRasterHeight()
name = product.getName()
desc = product.getDescription()
band_names = product.getBandNames()

print("Product: %s, %d x %d pixels, %s" % (name, width, height, desc))
print("Bands:   %s" % (band_names))

b7 = product.getBand('radiance_7')
b10 = product.getBand('radiance_10')
if b10 is None:
    b7 = product.getBand('M07_radiance')
    b10 = product.getBand('M10_radiance')
ndviProduct = Product('NDVI', 'NDVI', width, height)
ndviBand = ndviProduct.addBand('ndvi', ProductData.TYPE_FLOAT32)
ndviBand.setNoDataValue(numpy.nan)
ndviBand.setNoDataValueUsed(True)

writer = ProductIO.getProductWriter('BEAM-DIMAP')

ProductUtils.copyGeoCoding(product, ndviProduct)

ndviProduct.setProductWriter(writer)
ndviProduct.writeHeader(String('snappy_ndvi_with_masks_output.dim'))

r7  = numpy.zeros(width, dtype=numpy.float32)
r10 = numpy.zeros(width, dtype=numpy.float32)

v7  = numpy.zeros(width, dtype=numpy.uint8)
v10 = numpy.zeros(width, dtype=numpy.uint8)

print("Writing...")

for y in range(height):
    b7.readPixels(0, y, width, 1, r7)
    b10.readPixels(0, y, width, 1, r10)

    b7.readValidMask(0, y, width, 1, v7)
    b10.readValidMask(0, y, width, 1, v10)

    invalidMask7 = numpy.where(v7 == 0, 1, 0)
    invalidMask10 = numpy.where(v10 == 0, 1, 0)

    ma7 = numpy.ma.array(r7, mask=invalidMask7, fill_value=numpy.nan)
    ma10 = numpy.ma.array(r10, mask=invalidMask10, fill_value=numpy.nan)

    print("processing line ", y, " of ", height)
    ndvi = (ma10 - ma7) / (ma10 + ma7)
    ndviBand.writePixels(0, y, width, 1, ndvi.filled(numpy.nan))

ndviProduct.closeIO()

print("Done.")


import sys

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')
from esa_snappy import jpy

ProductIOPlugInManager = jpy.get_type('org.esa.snap.core.dataio.ProductIOPlugInManager')
ProductReaderPlugIn = jpy.get_type('org.esa.snap.core.dataio.ProductReaderPlugIn')
ProductWriterPlugIn = jpy.get_type('org.esa.snap.core.dataio.ProductWriterPlugIn')

read_plugins = ProductIOPlugInManager.getInstance().getAllReaderPlugIns()
write_plugins = ProductIOPlugInManager.getInstance().getAllWriterPlugIns()

print('Writer formats:')
while write_plugins.hasNext():
    plugin = write_plugins.next()
    print('  ', plugin.getFormatNames()[0], plugin.getDefaultFileExtensions()[0])

print(' ')

print('Reader formats:')
while read_plugins.hasNext():
    plugin = read_plugins.next()
    print('  ', plugin.getFormatNames()[0], plugin.getDefaultFileExtensions()[0])


import sys

# This example requires an arbitrary input product which can be read by SNAP

# This relative path is for the esa_snappy test setup only, as the esa_snappy module is located two folder levels above.
# Adapt this path if script shall be run from a different location!
sys.path.append('../../')
import esa_snappy
from esa_snappy import ProductIO, WKTReader

SubsetOp = esa_snappy.jpy.get_type('org.esa.snap.core.gpf.common.SubsetOp')

if len(sys.argv) != 3:
    print("usage: %s <file> <geometry-wkt>" % sys.argv[0])
    print("       %s ./TEST.N1 \"POLYGON((15.786082 45.30223, 11.798364 46.118263, 10.878688 43.61961, 14.722727"
          "42.85818, 15.786082 45.30223))\"" % sys.argv[0])
    sys.exit(1)

file = sys.argv[1]
wkt = sys.argv[2]

geom = WKTReader().read(wkt)

print("Reading...")
product = ProductIO.readProduct(file)

op = SubsetOp()
op.setSourceProduct(product)
op.setGeoRegion(geom)
op.setCopyMetadata(True)

sub_product = op.getTargetProduct()

print("Writing...")
ProductIO.writeProduct(sub_product, "snappy_subset_output.dim", "BEAM-DIMAP")

print("Done.")



from .binning_variable import BinningVariable
from .binning_variables import BinningVariables
from .output_bands import BinningOutputBands
from .binning_band import BinningBand
from .aggregators import Aggregators
